//———————————————————————————————————————————————————————————————————————————
// name: perpetual_motion.ck
// desc: HW3 Final Deliverable, a happier sequencer than last time.
//       Sequence plucked strings, wurlies, and kicks timed to the swing of each pendulum.
//
// Pendulum Physics reference: https://www.youtube.com/watch?v=uB6y9t4qt4I
//
// author: Elisse Chow
// date: Fall 2023
//———————————————————————————————————————————————————————————————————————————

GG.fullscreen();
GG.scene() @=> GScene @ scene;
GG.camera() @=> GCamera @ cam;
scene.light().intensity(0);  // disable default directional light

cam.posZ(45);
cam.fov(30);

// Modified from light.ck
class LightBulb extends GGen
{
    // GGen network. a light + sphere at the same position
    FlatMaterial mat;
    GPointLight light --> GSphere bulb --> this;
    GTorus torus_bulb;
    torus_bulb.sca(0.2);
    GCube cube_bulb;
    cube_bulb.sca(0.3);

    0 => int shape_chosen;

    fun void shape(int choice) {
        // if (shape_chosen == choice) return;

        if (shape_chosen == 0) {
            light --< bulb --< this;
        } else if (shape_chosen == 1) {
            light --< torus_bulb --< this;
        } else {
            light --< cube_bulb --< this;
        }

        if (choice == 0) {
            light --> bulb --> this;
        } else if (choice == 1) {
            light --> torus_bulb --> this;
        } else {
            light --> cube_bulb --> this;
        }
        choice => shape_chosen;
    }

    // set up sphere to be a flat color
    bulb.mat(mat);
    torus_bulb.mat(mat);
    cube_bulb.mat(mat);
    mat.color(@(1, 1, 1));
    @(0.4, 0.4, 0.4) => bulb.sca;

    // set light falloff
    light.falloff(0.045, 0.0075);  // falloff chart: https://wiki.ogre3d.org/tiki-index.php?page=-Point+Light+Attenuation

    vec3 lightCol;
    Math.random2f(0.5, 1.5) => float pulseRate;  // randomize pulse rate for fading in/out

    fun void color(float r, float g, float b) {
        @(r, g, b) => lightCol;  // save the set color
        mat.color(@(r, g, b));   // set material color
        light.diffuse(@(r, g, b));  // set light diffuse color
        light.specular(@(r, g, b));
    }

}

fun vec3 hsv_to_rgb(float h, float s, float v) {
    /*
        HSV to RGB converter.
        Thanks @tae kyu for the link to the calcs!
        https://www.rapidtables.com/convert/color/hsv-to-rgb.html

        Args:
            h: Hue [0, 360]
            s: Saturation [0, 1]
            v: Value [0, 1]
        Returns:
            vec3 for RGB color

    */
    v * s => float c;
    (h / 60) % 2 - 1 => float temp;
    if (temp < 0) {
        -temp => temp;
    }

    c * (1 - temp) => float x;
    v - c => float m;

    vec3 rgb_prime;
    if (0 <= h && h < 60) {
        @(c, x, 0.0) => rgb_prime;
    } else if (60 <= h && h < 120) {
        @(x, c, 0.0) => rgb_prime;
    } else if (120 <= h && h < 180) {
        @(0.0, c, x) => rgb_prime;
    } else if (180 <= h && h < 240) {
        @(0.0, x, c) => rgb_prime;
    } else if (240 <= h && h < 300) {
        @(x, 0.0, c) => rgb_prime;
    } else {
        @(c, 0.0, x) => rgb_prime;
    }

    vec3 rgb;
    (rgb_prime.x + m) => rgb.x;
    (rgb_prime.y + m) => rgb.y;
    (rgb_prime.z + m) => rgb.z;
    return rgb;
}

16 => int MAX_POS;  // MAXIMUM NUMBER OF NOTES
class Pendulum extends GGen {
    0 => static int NONE;       // Not selected or actively able to play.
    1 => static int ACTIVE;   // Pendulum in motion.
    2 => static int PLAYING;     // Making sound
    ACTIVE => int state;

    0 => static int DESELECTED;
    1 => static int SELECTED;
    DESELECTED => int input_state;

    // STANDARD INITIALIZATION
    float max_amplitude;
    float length;
    float g;

    vec3 anchor;

    float t;

    1.0 => float position_percent;

    LightBulb bulb;
    bulb.color(1, 1, 1);
    bulb --> this;

    GLines connection;
    connection.mat().color(@(0.5, 0.5, 0.5));
    connection --> this;

    PluckedString ps => Pan2 pan => dac;
    fun void set_freq(int freq) {
        freq => Std.mtof => ps.freq;
    }

    int notes[];
    int note_idx;

    fun void init(float max_amplitude, float length, float g, int notes[], int note_idx) {
        max_amplitude => this.max_amplitude;
        length => this.length;
        g => this.g;

        notes @=> this.notes;
        note_idx => this.note_idx;
        this.set_freq(this.notes[note_idx]);
    }

    fun void set_position(float percent) {
        if (percent <= 0.5) {
            0.5 => position_percent;
            return;
        } else if (percent > 1) {
            1 => position_percent;
            return;
        }
        percent => position_percent;
    }
    int instrument;
    fun void choose_instrument(int choice) {
        choice => instrument;
        bulb.shape(choice);
    }

    fun void set_length(float length) {
        length => this.length;
    }

    fun void color(vec3 c) {
        bulb.color(c.x, c.y, c.z);
    }

    max_amplitude => float prev_theta;

    Wurley wurl => pan => dac;
    Kick k => dac;

    // basic play function (add more arguments as needed)
    fun void wurl_play( float velocity )
    {
        Std.mtof( this.notes[this.note_idx] ) => wurl.freq;
        velocity => wurl.noteOn;
    }

    fun void update(float dt) {
        1.5 * dt +=> t;
        2 * Math.PI * Math.sqrt(length/g) => float period;
        max_amplitude * Math.cos(2 * Math.PI / period * t) => float theta;

        // Light intensity -> close to max_amplitude? brighter
        Math.pow(theta, 4) / Math.pow(max_amplitude, 4) + 0.1 => float intensity;
        
        calc_pend_pos(length, theta) => vec3 anchor_mid;
        calc_pend_pos(length, theta - 0.03) => vec3 anchor_l;
        calc_pend_pos(length, theta + 0.03) => vec3 anchor_r;
        connection.geo().positions([anchor, anchor_mid, anchor_l, anchor_mid, anchor_r]);

        0.8 * (MAX_POS - this.note_idx) / MAX_POS$float + 0.2 => position_percent;

        calc_pend_pos(position_percent * length, theta) => vec3 bulb_pos;

        bulb.pos(bulb_pos);
        
        if (state != ACTIVE && state != PLAYING) {
            if (input_state != SELECTED) {
                bulb.light.intensity(0.1);
                color_lerp(@(0.1, 0.1, 0.1), bulb.light.intensity() * bulb.lightCol, 0.1) => bulb.mat.color;
                return;
            } else {
                bulb.light.intensity(intensity);
                color_lerp(@(1, 1, 1), bulb.light.intensity() * bulb.lightCol, 0.3) => bulb.mat.color;
                return;
            }
        }

        Math.cos(2 * Math.PI / period * t) => pan.pan;
        // Set all to zero
        0 => ps.gain;
        0 => wurl.gain;
        0 => k.f.gain;
        // Start one.
        if (instrument == 0) {
            intensity => ps.gain;
        } else if (instrument == 1) {
            intensity => wurl.gain;
        } else {
            10 * intensity => k.f.gain;
        }

        bulb.light.intensity(intensity);
        color_lerp(@(0.1, 0.1, 0.1), bulb.light.intensity() * bulb.lightCol, intensity) => bulb.mat.color;

        if (state != PLAYING) {
            if (intensity >= 1 - 0.05) {
                bulb.light.falloff(0.045, 0.0075);
                this.set_freq(this.notes[this.note_idx]);
                if (instrument == 0) {
                    spork ~ ps.pluck(0.1::second, 0.25*intensity);
                } else if (instrument == 1) {
                    spork ~ wurl_play(0.25*intensity);
                } else {
                    bulb.light.falloff(0.35, 0.55);
                    color_lerp(@(0.1, 0.1, 0.1), bulb.light.intensity() * bulb.lightCol, intensity) => bulb.mat.color;
                    spork ~ k.play();
                }
                PLAYING => state;
            }
        }
        if (intensity < 1 - 0.05) {
            ACTIVE => state;
        }

        if (input_state == SELECTED) {
            bulb.light.intensity(intensity);
            color_lerp(@(1, 1, 1), bulb.light.intensity() * bulb.lightCol, 0.3) => bulb.mat.color;
        } 
    }

}

// Synthesized percussion instruments are borrowed (stolen!)
// from Tristan Peng's HW1: https://ccrma.stanford.edu/~pengt/256a/hw1.html

// thanks Tristan
class Kick extends Chugraph { 
    inlet => Noise n => LPF f => ADSR e => outlet;
    110 => f.freq;
    40 => f.gain;
    e.set(5::ms, 50::ms, 0.1, 100::ms);

    fun void play() {
        e.keyOn();
        50::ms => now;
        e.keyOff();
        e.releaseTime() => now;
    }
}

//  from chugraph.ck
class PluckedString extends Chugraph // Chubgraph
{
    // karplus + strong plucked string filter
    // Ge Wang (gewang@cs.princeton.edu)
    
    Noise imp => OneZero lowpass => outlet;
    lowpass => DelayA delay => lowpass;
    
    .99999 => float R;
    1/220 => float L;
    -1 => lowpass.zero;
    220 => freq;
    0 => imp.gain;
    
    fun float freq( float f )
    {
        // delay length
        1/f => L;
        // set delay length
        L::second => delay.delay;
        // set gain
        Math.pow( R, L ) => delay.gain;
        // return frequency through
        return f;
    }
    
    fun void pluck( dur ringDur, float gain )
    {
        // turn on the noise...
        gain => imp.gain;
        // for one delay length
        L::second => now;
        // turn off the noise
        0 => imp.gain;
        // let it ring
        ringDur => now;
    }
}

fun vec3 calc_pend_pos(float length, float theta) {
    vec3 pos;
    length * Math.sin(theta) => pos.x;
    0 - length * (Math.cos(theta)) => pos.y;
    return pos;
}

fun vec3 color_lerp(vec3 c0, vec3 c1, float percentage) {
    return (1.0 - percentage) * c0 + (percentage) * c1;
}

8 => int NUM_PENDULUM;
Pendulum p[NUM_PENDULUM];
0 => int num_shown;  // Empty black to begin

for (int i; i < NUM_PENDULUM; i++) {
    p[i].init(Math.PI * 2 / 3, 0.5*i+5, 9.81, [55, 57, 59, 60, 62, 64, 65, 67, 69, 71, 72, 74, 76, 77, 79, 81], 8);
    hsv_to_rgb(i*(360/NUM_PENDULUM$float), 1, 1) => vec3 color;
    p[i].color(color);
    if (i < num_shown) {
        p[i] --> scene;
    }
}
1 => p[0].input_state;  // Default have the first pendulum selected.

// Ground for lights to cast on
GPlane ground --> GG.scene();
@(100, 100, 1) => ground.sca;
ground.mat().color(@(0, 0, 0));
ground.posZ(-3);

int SELECTED_BUDDY;
int equal_down, minus_down, s_down, next_down, o_down, j_down, k_down, backslash_down, backspace_down;
int one_down, two_down, three_down;  // Instruments

spork ~ kb_edit_listener();
spork ~ kb_instrument_listener();
fun void kb_instrument_listener() {
    while(true) {
        p[SELECTED_BUDDY] @=> Pendulum @ p_temp;

        if ( KB.isKeyDown(KB.KEY_1) ) {
            if (one_down == 0) {
                1 => one_down;
                p_temp.choose_instrument(0);
            }
        } else {
            0 => one_down;
        }

        if ( KB.isKeyDown(KB.KEY_2) ) {
            if (two_down == 0) {
                1 => two_down;
                p_temp.choose_instrument(1);
            }
        } else {
            0 => two_down;
        }

        if ( KB.isKeyDown(KB.KEY_3) ) {
            if (three_down == 0) {
                1 => three_down;
                p_temp.choose_instrument(2);
            }
        } else {
            0 => three_down;
        }
        16::ms => now;
    }
}
fun void kb_edit_listener() {
    while( true )
    {
        p[SELECTED_BUDDY] @=> Pendulum @ p_temp;
        // Select the next pendulum
        if( KB.isKeyDown(KB.KEY_N) ) {
            if (next_down == 0) {
                1 => next_down;
                0 => p_temp.input_state;
                ( SELECTED_BUDDY + 1 ) % num_shown => SELECTED_BUDDY;
                1 => p[SELECTED_BUDDY].input_state;
            }
        } else {
            0 => next_down;
        }

        if (SELECTED_BUDDY < num_shown) {
            // Toggle the plucking of the pendulum sound
            if ( KB.isKeyDown(KB.KEY_O) ) {
                if (o_down == 0) {
                    1 => o_down;
                    if (p_temp.state == p_temp.ACTIVE) {
                        p_temp.NONE => p_temp.state;
                    } else if (p_temp.state == p_temp.NONE) {
                        p_temp.ACTIVE => p_temp.state;
                    }
                }
            } else {
                0 => o_down;
            }

            // Move ball down on the pendulum
            if ( KB.isKeyDown(KB.KEY_J) ) {
                if (j_down == 0) {
                    1 => j_down;
                    1 -=> p_temp.note_idx;
                    if (p_temp.note_idx < 0) 0 => p_temp.note_idx;
                }
            } else {
                0 => j_down;
            }
            // Move ball up on the pendulum
            if ( KB.isKeyDown(KB.KEY_K) ) {
                if (k_down == 0) {
                    1 => k_down;
                    1 +=> p_temp.note_idx;
                    if (p_temp.note_idx > MAX_POS - 1) MAX_POS - 1 => p_temp.note_idx;
                }
            } else {
                0 => k_down;
            }
        }

        // Show more of the pendulums in the scene
        if ( KB.isKeyDown(KB.KEY_EQUAL) ) {
            if (equal_down == 0) {
                1 => equal_down;
                if (num_shown < NUM_PENDULUM) {
                    if (num_shown == 0) {
                        0 => p[0].t;
                    }
                    p[0].t => p[num_shown].t;
                    p[num_shown] --> scene;
                    num_shown++;
                }
            }
        } else {
            0 => equal_down;
        }

        // Hide pendulums in the scene
        if ( KB.isKeyDown(KB.KEY_MINUS) ) {
            if (minus_down == 0) {
                1 => minus_down;
                if (num_shown > 0) {
                    0 => p[num_shown - 1].ps.gain;
                    p[num_shown - 1] --< scene;

                    num_shown--;
                }
            }
        } else {
            0 => minus_down;
        }

        if ( KB.isKeyDown(KB.KEY_BACKSLASH) ) {
            if (backslash_down == 0) {
                1 => backslash_down;

                for (int i; i < NUM_PENDULUM; i++) {
                    if (i < num_shown) p[i] --< scene;
                    0 => p[i].t;
                    p[i].DESELECTED => p[i].input_state;
                    p[i].choose_instrument(0);
                    p[i].ACTIVE => p[i].state;
                    i => p[i].note_idx;
                    p[i] --> scene;
                }
                p[0].SELECTED => p[0].input_state;
                0 => SELECTED_BUDDY;
                NUM_PENDULUM => num_shown;
            }
        } else {
            0 => backslash_down;
        }

        if ( KB.isKeyDown(KB.KEY_BACKSPACE) ) {
            if (backspace_down == 0) {
                1 => backspace_down;
                for (int i; i < num_shown; i++) {
                    0 => p[i].ps.gain;
                    8 => p[i].note_idx;
                    p[i] --< scene;
                }
                0 => num_shown;
                p[0].SELECTED => p[0].input_state;
            }
        } else {
            0 => backspace_down;
        }
        // every so often!
        16::ms => now;
    }
}

while (true) {
    GG.nextFrame() => now;
}